import pandas as pd
import zipfile


weirds = ("indp02 indp07 occp02 occp10 OCCP10 OCCP12 "
          "socp00 socp10 naicsp02 naicsp07".split())
adj_cols = "INTP OIP PAP PERNP PINCP RETP SEMP SSIP SSP WAGP".split()
def read_chunks(fname, chunksize=10**5, voters_only=False, adj_inc=None):
    chunks = pd.read_csv(
        fname,
        skipinitialspace=True,
        na_values={k: ['N.A.', 'N.A.//', 'N.A.////'] for k in weirds},
        dtype={k: str for k in weirds},
        chunksize=chunksize,
    )
    for chunk in chunks:
        if voters_only:
            chunk = chunk[(chunk.AGEP >= 18) & (chunk.CIT != 5)]
        
        if adj_inc is None:
            if 'ADJINC' in chunk:
                adj_inc = True
            elif 'ADJINC_orig' in chunk:
                adj_inc = False
            else:
                raise ValueError("Unclear whether income has been adjusted, "
                                 "and adj_inc is None; pass either True or "
                                 "False explicitly")

        if adj_inc:
            adj = chunk.ADJINC / 1e6
            for k in adj_cols:
                chunk[k] *= adj
            chunk['ADJINC_orig'] = chunk.ADJINC
            del chunk['ADJINC']

        yield chunk


VERSIONS = {
    '2006-10': {
        'weight_cols': ['PWGTP'] + ['PWGTP{}'.format(i) for i in range(1, 81)],
        'meta_cols': "RT SPORDER serialno PUMA ST".split(),
        'discrete_feats': '''CIT COW ENG FER GCL GCM GCR JWRIP JWTR LANX MAR
                             MIG MIL MLPA MLPB MLPC MLPD MLPE MLPF MLPG MLPH
                             MLPI MLPJ MLPK NWAB NWAV NWLA NWLK NWRE RELP SCH
                             SCHG SCHL SEX WKL WKW ANC ANC1P ANC2P DECADE
                             DRIVESP ESP ESR HISP indp02 indp07 LANP MIGPUMA
                             MIGSP MSP naicsp02 naicsp07 NATIVITY NOP OC occp02 
                             occp10 PAOC POBP POWPUMA POWSP QTRBIR RAC1P RAC2P
                             RAC3P RACAIAN RACASN RACBLK RACNHPI RACSOR RACWHT
                             RC SFN SFR socp00 socp10 VPS WAOB'''.split(),
        'alloc_flags': '''FAGEP FANCP FCITP FCOWP FENGP FESRP FFERP FGCLP
                          FGCMP FGCRP FHISP FINDP FINTP FJWDP FJWMNP FJWRIP
                          FJWTRP FLANP FLANXP FMARP FMIGP FMIGSP FMILPP FMILSP
                          FOCCP FOIP FPAP FPOBP FPOWSP FRACP FRELP FRETP FSCHGP
                          FSCHLP FSCHP FSEMP FSEXP FSSIP FSSP FWAGP FWKHP FWKLP
                          FWKWP FYOEP'''.split(),
        'real_feats': '''AGEP INTP JWMNP OIP PAP RETP SEMP SSIP SSP WAGP WKHP
                         YOEP JWAP JWDP PERNP PINCP POVPIP RACNUM'''.split(),
        'region_year': '00',
    }
}

# feat_cols = set(discrete_feats) | set(alloc_flags) | set(real_feats)
# all_cols = set(meta_cols) | feat_cols



from __future__ import division, print_function
from collections import Counter
import heapq
import os
import sys
import zipfile

import numpy as np
import pandas as pd
import progressbar as pb  # should be progressbar2

from .data import geocode_data
from .reader import read_chunks, VERSIONS


def sort_by_region(source, out_fmt, voters_only=True, adj_inc=True,
                   version='2006-10', chunksize=10**5, n_to_sample=5000,
                   stats_only=False):
    info = VERSIONS[version]
    all_cols = set(info['meta_cols'] + info['weight_cols'] + info['real_feats']
                   + info['discrete_feats'] + info['alloc_flags'])
    real_feats = info['real_feats']
    discrete = info['discrete_feats'] + info['alloc_flags']

    key = 'puma_region_{}'.format(info['region_year'])
    puma_to_region = geocode_data(key).region.to_dict()

    created_files = set()

    real_info = []  # (num non-nan, mean series, mean of square series) entries
    value_counts = {}  # column name => sum of col.value_counts()

    n_total = 0
    not_in_region = Counter()

    if isinstance(source, list):
        files = source
        opener = open
    else:
        z = zipfile.ZipFile(source, 'r')  # close through GC :|
        opener = z.open
        files = [fn for fn in z.namelist() if fn.endswith('.csv')]

    # We're going to use weighted reservoir sampling to keep a few random rows
    # in memory, so we can estimate the median pairwise distance between
    # features later.
    reservoir = []

    for file in files:
        print("File {}".format(file), file=sys.stderr)
        bar = pb.ProgressBar(max_value=pb.UnknownLength)
        read = 0
        bar.update(read)

        checked_cols = False
        with opener(file, 'r') as f:
            for chunk in read_chunks(f, voters_only=voters_only,
                                     adj_inc=adj_inc, chunksize=chunksize):
                if not checked_cols:
                    cols = set(chunk.columns)
                    assert (cols - all_cols).issubset({'ADJINC', 'ADJINC_orig'})
                    assert not (all_cols - cols)
                    checked_cols = True
                read += chunk.shape[0]

                # components of mean / std for real-valued features
                reals = chunk[real_feats]
                real_info.append((
                    reals.shape[0] - reals.isnull().sum(),
                    reals.mean(),
                    (reals.astype(np.float128) ** 2).mean()))

                # add onto value counts for discrete features
                for k in discrete:
                    value_counts[k] = chunk[k].value_counts().add(
                        value_counts.get(k, 0), fill_value=0)

                # manage reservoir sample
                rs = np.asarray(
                    np.random.uniform(size=chunk.shape[0]) ** (1 / chunk.PWGTP))
                for r_tup in zip(rs, chunk.itertuples()):
                    # TODO: could speed this up if it's slow, probably
# maybe there's a weighted version of
# http://erikerlandson.github.io/blog/2015/11/20/very-fast-reservoir-sampling/
                    if len(reservoir) < n_to_sample:
                        heapq.heappush(reservoir, r_tup)
                    else:
                        heapq.heappushpop(reservoir, r_tup)

                # output into files by region
                if not stats_only:
                    regions = np.empty(chunk.shape[0], dtype=object)
                    for i, tup in enumerate(zip(chunk.ST, chunk.PUMA)):
                        regions[i] = r = puma_to_region.get(tup)
                        if r is None:
                            not_in_region[tup] += 1

                    for r, r_chunk in chunk.groupby(regions):
                        out = out_fmt.format(r)
                        try:
                            r_chunk.to_hdf(
                                out, 'df', format='table', append=True,
                                mode='a' if r in created_files else 'w',
                                complib='blosc', complevel=6)
                        except ValueError:
                            # if new chunk has longer strings than previous
                            # one did, this will cause an error...instead of
                            # hardcoding sizes, though, just re-write the data
                            old = pd.read_hdf(out, 'df')
                            new = pd.concat([old, r_chunk])
                            r_chunk.to_hdf(
                                out, 'df', format='table', mode='w',
                                complib='blosc', complevel=6)
                        created_files.add(r)
                bar.update(read)
        bar.finish()
        n_total += read

    if not_in_region:
        print("Records not in a region:")
        for k, count in not_in_region.most_common():
            print(k, count)

    total = sum(n_nonnan for n_nonnan, mean, mean_sq in real_info)
    real_means = 0
    real_mean_sqs = 0
    for n_nonnan, mean, mean_sq in real_info:
        real_means += n_nonnan / total * mean
        real_mean_sqs += n_nonnan / total * mean_sq

    real_stds = (real_mean_sqs - real_means.astype(np.float128)**2
        ).apply(np.sqrt).astype(np.float64)

    stats = {}
    stats['real_means'] = real_means
    stats['real_stds'] = real_stds
    stats['value_counts'] = value_counts
    stats['n_total'] = n_total
    stats['version'] = version
    stats['sample'] = pd.DataFrame([t for r, t in reservoir])
    return stats



import pandas as pd
import six

def save_stats(fn, stats):
    with pd.HDFStore(fn, 'w') as f:
        stats['sample'].to_hdf(f, 'sample', format='table')
        stats['real_means'].to_hdf(f, 'real_means')
        stats['real_stds'].to_hdf(f, 'real_stds')
        for k, v in six.iteritems(stats['value_counts']):
            v.to_hdf(f, 'value_counts/{}'.format(k))
        for k in ['n_total', 'version']:
            pd.Series([stats[k]]).to_hdf(f, k)


def load_stats(fn):
    stats = {}
    with pd.HDFStore(fn, 'r') as f:
        stats['sample'] = f['sample']
        stats['real_means'] = f['real_means']
        stats['real_stds'] = f['real_stds']
        stats['n_total'] = f['n_total'].iloc[0]
        stats['version'] = f['version'].iloc[0]

        stats['value_counts'] = v = {}
        pre = '/value_counts/'
        for k in f.keys():
            if k.startswith(pre):
                v[k[len(pre):]] = f[k]
    return stats



from __future__ import division, print_function
import sys

import numpy as np
import pandas as pd
import progressbar as pb
from sklearn.metrics.pairwise import euclidean_distances
import six

from .reader import VERSIONS, read_chunks


def get_dummies(df, stats, num_feats=None, ret_df=True, dtype=np.float64,
                out=None):
    '''
    Gets features for the person records in `df`: standardizes the real-valued
    features, and does one-hot encoding for the discrete ones.
    '''
    info = VERSIONS[stats['version']]
    if num_feats is None:
        num_feats = _num_feats(stats)

    if out is None:
        out = np.empty((df.shape[0], num_feats), dtype=dtype)
    else:
        assert out.shape == (df.shape[0], num_feats)

    reals = out[:, :len(info['real_feats'])]
    reals[:] = df[info['real_feats']]
    reals[:] -= stats['real_means']
    reals[:] /= stats['real_stds']
    reals[np.isnan(reals)] = 0
    if ret_df:
        feat_names = list(info['real_feats'])
    start_col = len(info['real_feats'])

    for k in info['discrete_feats'] + info['alloc_flags']:
        vc = stats['value_counts'][k]
        c = pd.Categorical(df[k], categories=vc.index).codes
        n_codes = len(vc)
        if ret_df:
            feat_names += ['{}_{}'.format(k, v) for v in vc.index]
        if vc.sum() < stats['n_total']:
            c = c.copy()
            c[c == -1] = n_codes
            n_codes += 1
            if ret_df:
                feat_names.append('{}_nan'.format(k))
        bit = out[:, start_col:start_col + n_codes]
        np.eye(n_codes).take(c, axis=0, out=bit)
        start_col += n_codes
    assert start_col == num_feats

    if ret_df:
        return pd.DataFrame(out, index=df.index, columns=feat_names)
    else:
        return out


def _num_feats(stats):
    n = stats['real_means'].size
    n_total = stats['n_total']
    for k, v in six.iteritems(stats['value_counts']):
        n += v.size + (1 if v.sum() < n_total else 0)
    return n


################################################################################
### Embeddings

def linear_embedding(feats, wts, out=None):
    '''
    Gets the linear kernel embedding (which is just the weighted mean) for
    dummy features `feats`, with sample weighting `wts`.
    '''
    if out is None:
        out = np.empty(feats.shape[1])
    np.dot(wts, feats, out=out)
    out /= wts.sum()
    return out


def rff_embedding(feats, wts, freqs, out=None):
    '''
    Gets the random Fourier feature embedding for dummy features `feats`,
    with sample weighting `wts`.
    '''
    D = freqs.shape[1]
    if out is None:
        out = np.empty(2 * D)

    angles = np.dot(feats, freqs)
    sin_angles = np.sin(angles)  # TODO: could use MKL sincos for this
    cos_angles = np.cos(angles, out=angles)

    np.dot(wts, sin_angles, out=out[:D])
    np.dot(wts, cos_angles, out=out[D:])
    out /= wts.sum()
    return out


def pick_rff_freqs(n_freqs, bandwidth, stats=None, n_feats=None):
    '''
    Sets up sampling with random Fourier features corresponding to a Gaussian
    kernel with the given bandwidth, with an embedding dimension of `2*n_freqs`.
    '''
    if n_feats is None:
        n_feats = _num_feats(stats)
    return np.random.normal(0, 1 / bandwidth, size=(n_feats, n_freqs))


def pick_gaussian_bandwidth(stats):
    '''
    Finds the median distance between features from the random sample saved
    in stats.
    '''
    samp = get_dummies(stats['sample'], stats, ret_df=False)
    D2 = euclidean_distances(samp, squared=True)
    return np.sqrt(np.median(D2[np.triu_indices_from(D2, k=1)]))


################################################################################

def get_embeddings(files, stats, n_freqs=2048, freqs=None, bandwidth=None,
                   chunksize=2**13, skip_rbf=False):
    n_feats = _num_feats(stats)
    feat_names = None

    if not skip_rbf:
        if freqs is None:
            if bandwidth is None:
                print("Picking bandwidth by median heuristic...",
                      file=sys.stderr, end='')
                bandwidth = pick_gaussian_bandwidth(stats)
                print("picked {}".format(bandwidth), file=sys.stderr)
            freqs = pick_rff_freqs(n_freqs, bandwidth, n_feats=n_feats)
        else:
            n_freqs = freqs.shape[1]

    emb_lin = np.empty((len(files), n_feats))
    if not skip_rbf:
        emb_rff = np.empty((len(files), 2 * n_freqs))

    bar = pb.ProgressBar(max_value=stats['n_total'])
    bar.start()
    read = 0
    dummies = np.empty((chunksize, n_feats))
    for i, file in enumerate(files):
        lin_emb_pieces = []
        if not skip_rbf:
            rff_emb_pieces = []
        weights = []
        total_weight = 0
        for c in pd.read_hdf(file, chunksize=chunksize):
            feats = dummies[:c.shape[0], :]

            if feat_names is None:
                df = get_dummies(c, stats, num_feats=n_feats, ret_df=True,
                                 out=feats)
                feat_names = list(df.columns)
            else:
                get_dummies(c, stats, num_feats=n_feats, ret_df=False,
                            out=feats)

            lin_emb_pieces.append(linear_embedding(feats, c.PWGTP))
            if not skip_rbf:
                rff_emb_pieces.append(rff_embedding(feats, c.PWGTP, freqs))
            w = c.PWGTP.sum()
            weights.append(w)
            total_weight += w

            read += c.shape[0]
            bar.update(read)

        emb_lin[i] = 0
        for w, l in zip(weights, lin_emb_pieces):
            emb_lin[i] += l * (w / total_weight)

        if not skip_rbf:
            emb_rff[i] = 0
            for w, r in zip(weights, rff_emb_pieces):
                emb_rff[i] += r * (w / total_weight)
    bar.finish()

    if skip_rbf:
        return emb_lin, feat_names
    else:
        return emb_lin, emb_rff, freqs, bandwidth, feat_names
